function [v, errest, t, errests, hfinal] = SDIRK_43_adaptive(t0, tout, M, A, u0, TOL, hstart)
% Approximate the solution to
%    M \dot{x} = Ax, x(0) = u0
% at t = tout by a 4th-order SDIRK method with embedded 3rd-order error
% estimation. The method coefficients are taken from [1, p.100].
% 
% Returns the approximations vout, the error estimate errest, the vector of
% used time steps t, the error estimates at these time steps errests, and
% the final suggested step size hfinal. (The latter is not necessarily the
% final step size, since it will likely be reduced in order to exactly hit
% tend.)
% 
% If size(u0) ==  [n, r] and length(tout) = Nout then 
% size(vout) == [n, Nout*r], i.e. the first r columns constitute the
% approximation at the first output point, the next r columns the
% approximation at the second output point, and so on.
%
% [1] Hairer, E. and Wanner, G., Solving Ordinary Differential Equations
% II: Stiff and Differential-Algebraic Problems, 2nd Ed., Springer, 2002.

% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.

    if nargin < 5
        TOL = 1e-5;
    end
    
    [n, p] = size(u0);
    errest = Inf;
    errest_order = 3;
%     errest_order = errest_order - 1; % EPUS
%     TOL = tend * TOL;

    c = [1/4, 3/4, 11/20, 1/2, 1];
    aij= [     1/4,         0,      0,      0,   0; 
               1/2,       1/4,      0,      0,   0; 
             17/50,     -1/25,    1/4,      0,   0;
          371/1360, -137/2720, 15/544,    1/4,   0; 
             25/24,    -49/48, 125/16, -85/12, 1/4];
	beta = [25/24, -49/48, 125/16, -85/12, 1/4]; 
    beta_err = [-3/16, -27/32, 25/32, 0, 1/4];  
    
    onep = ones(1, p);
    B = zeros(5*p, 2*p);
    for j = 1:5
        B((j-1)*p+1:j*p, 1:p) =  diag(beta(j)*onep);
        B((j-1)*p+1:j*p, p+1:2*p) =  diag(beta_err(j)*onep);
    end
    
    C = zeros(5*p, 5*p);
    for j = 1:5
        for i = 1:j-1
         C((i-1)*p+1:i*p, (j-1)*p+1:j*p) = diag(aij(j,i)*onep);
        end
        C((j-1)*p+1:j*p, 1:p) =  diag(beta(j)*onep);
    end

    t = t0;
    Nout = length(tout);
    if nargin < 7
        hj = (tout(1)-t0) / 100;
    else
        hj = min(hstart, tout(1)-t0);
        if hj > 0 % if for some reason the user asked for t0 = 0 and tout(1) = 0 we'll have a problem further down the line
            hfinal = hj;
        else
            hfinal = hstart;
        end
    end
        
    j = 1; % internal step
    u = u0;
    vout = [];
    for k = 1:Nout % for all output times
        tk = tout(k); % next output
        hj = min(hj, tk-t(end));
        while t(end) + hj <= tk
            K = zeros(n, 5*p);

            for i = 1:5
                  K(:, (i-1)*p+1:i*p) = (M- aij(i,i)*hj*A)\(A*(u + (hj*K(:, 1:(i-1)*p)*C(1:(i-1)*p, (i-1)*p+1:i*p))));
            end

            % 4th-order approximation
    %         v2 = u + hj*(25/24*k1 - 49/48*k2 + 125/16*k3 - 85/12*k4 + 1/4*k5);  
            v = u + hj*K*B(:,1:p);
            % 3rd-order approximation
        %     v = u + h*(59/48*k1 - 17/96*k2 + 225/32*k3 - 85/12*k4);  
            % Difference between 4th- and 3rd-order approximations, evaluated
            % in a better way
    %         errest = hj*norm( -3/16*k1 -27/32*k2 + 25/32*k3 + 0*k4 + 1/4*k5);  
            errest = hj*norm(K*B(:,p+1:2*p));  

    %         errest = errest / hj; % EPUS

            if errest > TOL % redo step
                hj = (0.9*TOL/errest)^(1/(errest_order)) * hj;
            else
                t(end+1) = t(end) + hj;
                errests(j) = errest;

                % New step size
                hj = (TOL/errest)^(1/errest_order) * hj;

                j = j+1;
                u = v;

                if abs(t(end) - tk) < eps % We are at the next output time
                    vout = [vout, v];
                    break
                end

                if t(end) + hj > tk % Ensure that we end up precisely at tk
                    hfinal = hj;
                    hj = tk-t(end);
                end
            end
        end
    end
end