function [t, Ls, Ds, ms, time, timeIQ] = Strang(args)
% Approximate the solution to 
%
% \dot(P) = A'P + PA + Q - PB Rinv B'P  , P(0) = P0,
%
% by taking Nt time steps of size h with the Strang splitting scheme.
%
% Here,
% P0 = L0 D0 L0' and Q = IQL IQD IQL'
%
% All inputs submitted in argument structure args. 
%
% Outputs: t, time steps
%          Ls, low-rank factors L at time steps (if args.intermediates)
%          otherwise only at final time
%          Ds, as above, but for D in LDL^T
%          ms, as above, but for the ranks of L
%          time, estimated total elapsed computation time, excluding
%                unpacking and trivial setup
%          timeIQ, estimated computation time for evaluating the
%                  integral IQ(h/2)

%
% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.
%

    % Unpack most of args for clarity
    A = args.A;  B = args.B; Rinv = args.Rinv;  M = args.M;
    LQ = args.LQ; DQ = args.DQ; L0 = args.L0; D0 = args.D0;    
    t0 = args.t0;  tend = args.tend;  Nt = args.Nt;
    intermediates = args.intermediates; print_step = args.print_step; 
    
    % Check for time dependency of A, B, C, M
    if ~isfield(args.impODE,'method') || (isfield(args.impODE,'method') && strcmp(args.impODE.method, 'timevarying'))
        timevarying = true;
        args.timevarying = true;
    else
        timevarying = false;
        args.timevarying = false;
    end
    
    if ~isfield(args, 'quadrature') || ~isfield(args.quadrature, 'order')
        % Since this splitting scheme is order 2, this should be enough.
        % 3 is typically not enough, since "order" here refers to the
        % actual error over [0, h]. In the time integration context, the
        % "order" refers to the error over the whole time interval and one 
        % extra power of h is gained on [0, h].
        % Thus using quadrature order 3 would give an error over [0, h]
        % of the same size as the local error of the splitting scheme, 
        % while we would like it to be negligible in comparison.
        args.quadrature.order = 5;
    end
    
    h = (tend-t0) / Nt;
    t = linspace(t0, tend, Nt+1);    
        
    ms = zeros(Nt, 1);
    ms(1) = size(L0,2);
    
    if intermediates
        Ls{Nt+1} = [];
        Ls{1} = L0;
        Ds{Nt+1} = [];
        Ds{1} = D0;
    else
        Ls = L0;
        Ds = D0;
    end
    
    args.impODEobj = impODEobj(args, true);

    telapsed = tic;

    
    % If not time-varying, we may approximate the integral once and for all
    if ~timevarying
        if args.print_output
            fprintf('Computing low-rank factorization of integral I_Q. \n');
        end
        timeIQ = tic;
        [IQL, IQD] = compute_integral_factors_LDLT(LQ, DQ, h, 1/2, args);
        timeIQ = toc(timeIQ);
    else
        timeIQ = 0;
    end
   
    if args.print_output
        fprintf('Starting time-stepping. \n');
    end
    
    for j = 1:Nt
        if print_step
            fprintf('Step: %d of %d. (Strang) \n', j, Nt);
        end
        
        % If time-varying, we have to compute a new integral term in each
        % step.
        if timevarying
            tIQ = tic;
            [IQL, IQD] = compute_integral_factors_LDLT(LQ, DQ, h, 1/2, args, [t(j), t(j) + h/2]);
            timeIQ = timeIQ + toc(tIQ);
        end
        
        if intermediates
            [Ls{j+1}, Ds{j+1}] = Strang_step(h, B, IQL, IQD, Rinv, Ls{j}, Ds{j}, args, t(j));
            ms(j+1) = size(Ls{j+1},2);
        else
            [Ls, Ds] = Strang_step(h, B, IQL, IQD, Rinv, Ls, Ds, args, t(j));
            ms(j+1) = size(Ls,2);
        end
        
        if args.print_output
            fprintf('Size of solution low-rank factor: %d \n', ms(j+1));
        end
        
    end
    time = toc(telapsed);

end