function [xis_rep, m, nbest] = find_best_poles(T0, T1, TOL)
% Find p repeated poles xis_rep and total number of rational Krylov
% iterations m = nbest(p-1) + 1 which yields an error of less than TOL when 
% approximating exp(tau*A)*b, for all tau in [T0, T1]. This should hold for
% all symmetric negative semi-definite operators A. The method uses the
% procedure outlined in [1, p.113-114] iterating over polynomials of degree
% n = 1:30.
% 
% [1] G{\"u}ttel, Stefan, Rational {K}rylov Methods for Operator Functions,
% Dissertation, 2010, Technische Universit{\"a}t Bergakademie Freiberg, 
% http://eprints.ma.man.ac.uk/2586/.
% 
% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.

    % Precomputed discretization of the function eta_n(tau):
    e = load('etas');
    etas = e.etas;
    ntaus = e.ntaus;
    taus = e.taus;
    nmax = size(etas, 1);
    
    m = Inf;
    nbest = Inf;
    for n = 1:nmax
       [xis_n, m_n] = find_xis(n);
       if m_n < m
           m = m_n;
           xis_rep = xis_n;
           nbest = n;
       end
    end

        %%% Nested function
        function [xis_rep, m] = find_xis(n)
        % Do the above procedure for the polynomial of degree n

        %%% Find the longest interval where metas < TOL:
        mmin = etas(n, :) < TOL;
        % Skip tau=0 because of singularity:
        istart = 2; 
        maxlen = -inf;

        while true
            iend = find(mmin(istart:end) == 0, 1, 'first') + istart-2;
            if  isempty(iend) || (iend > ntaus)
                if mmin(end) == 1
                    iend = ntaus;
                else
                    break
                end
            end
            newlen = iend - istart + 1;
            if newlen > maxlen
                maxlen = newlen;
                mstart = istart;
                mend = iend;        
            end
            istart = find(mmin(iend+1:end) == 1, 1, 'first') + iend;
            if isempty(istart)
                break;
            end
        end

        maxint = taus(mstart:mend);
        mint0 = taus(mstart);
        mint1 = taus(mend);
        % Convert from indices to actual length
        c = mint1 - mint0;

    %     % illustrate it
    %     semilogy(taus, etas(n, :))
    %     line([mint0, mint1], [TOL, TOL])

        if c > 1
            % This implies that the union of [T0*c^0, T0*c^1], [T0*c^1, T0*c^2], ...,
            % [T0*c^(p-1), T0*c^p] contains [T0, T1]:
            p = ceil(log(T1/T0) / log(c));
        elseif c > 0
            % In this case, 0 < c < 1 and we can cover the interval using the 
            % same idea but starting from T1 instead
            p = ceil(log(T0/T1) / log(c));
        end

        xi0 = 1;
        if c > 0 
            % Rescale xis
            xi1 = xi0 * mint0 / T0;

            xis_rep = xi1 ./ c.^(0:p-1);

            m = p*(n-1) + 1;
        else % We couldn't find any interval where the function is < TOL 
            m = Inf;
            xis_rep = [];
        end

    end
    
end