classdef impODEobj < handle

%
% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.
%

    properties
        transpose
        M
        dM % derivative of M, for the timevarying case
        A
        fun
        keys
        data
        IRKdata % data for the IRK methods
        Krylovdata % data for the (block) Krylov method
        t % current time, for the timevarying case
    end
    methods
        function obj = impODEobj(args, transpose)
            
            if nargin < 2
                obj.transpose = false;
            else
                obj.transpose = transpose;
            end
            
            if ~isfield(args, 'impODE')
                error('Must define impODE structure with parameters for the computation of matrix exponential actions!')
            end
            
            if ~isfield(args.impODE, 'method')
                method = 'IRK';
            else
                method = args.impODE.method; 
            end
            
            if strcmp(method, 'timevarying')
                timevarying = true;
            else
                timevarying = false;
            end
            if obj.transpose
                if timevarying
                    obj.M = @(t) args.M(t)';
                    obj.dM = @(t) args.dM(t)';
                    obj.A = @(t) args.A(t)';
                else
                    obj.M = args.M';
%                     obj.dM = args.dM;
                    obj.A = args.A';
                end
            else
                obj.M = args.M;
                if timevarying
                    obj.dM = args.dM;
                end
                obj.A = args.A;
            end
            
            obj.keys = [];
            obj.data = {};
                
            switch method
                % Time-varying! Only choice currently the adaptive
                % SDIRK43 method
                case 'timevarying'
                    obj.fun = @obj.timevarying;
                    
                    if ~isfield(args.impODE, 'tol')
                        obj.IRKdata.tol = 1e-5;
                    else
                        obj.IRKdata.tol = args.impODE.tol;
                    end
                    
                    if ~isfield(args.impODE, 'hstart')
                        % We have no idea about the time scale, but it
                        % seems reasonable that the total integration is
                        % over [0, O(1)], so try a step which is 1/100 of
                        % that
                        obj.IRKdata.hstart = 1e-2; 
                    else
                        obj.IRKdata.hstart = args.impODE.hstart;
                    end
                    
                    obj.IRKdata.fun = @SDIRK_43_adaptive_timevarying;
                    
                    
                % Adaptive IRK
                case 'IRKadaptive'
                    obj.fun = @obj.IRKadaptive;
                    
                    if ~isfield(args.impODE, 'tol')
                        obj.IRKdata.tol = 1e-5;
                    else
                        obj.IRKdata.tol = args.impODE.tol;
                    end
                    
                    obj.IRKdata.fun = @SDIRK_43_adaptive;
                
                % IRK
                case 'IRK'
                    obj.fun = @obj.IRK;
                    
                    if ~isfield(args.impODE, 'IRKorder')
                        obj.IRKdata.order = 5;
                    else
                        obj.IRKdata.order = args.impODE.IRKorder;
                    end
                    
                    if ~isfield(args.impODE, 'iterate')
                        obj.IRKdata.iterate = false;
                    else
                        obj.IRKdata.iterate = args.impODE.iterate;
                    end
                    
                    if obj.IRKdata.iterate
                        if ~isfield(args.impODE, 'tol')
                            obj.IRKdata.tol = 1e-6;
                        else
                            obj.IRKdata.tol = args.impODE.tol;
                        end
                    end
                    
                    if ~isfield(args.impODE, 'disk_cache')
                        obj.IRKdata.disk_cache = false;                        
                    else
                        obj.IRKdata.disk_cache = args.impODE.disk_cache;
                    end
                    
                    if obj.IRKdata.disk_cache
                        if ~isfield(args.impODE, 'cachepath')
                            obj.IRKdata.cachepath = './';                        
                        else
                            obj.IRKdata.cachepath = args.impODE.cachepath;
                        end
                    end
                    
                    switch obj.IRKdata.order
                        case 8
                            obj.IRKdata.fun = @impODE_GaussLegendre8; % TODO: coefficients got lost?
                            obj.IRKdata.jacfun = @gauss8_jacobian;
                        case 6
                            obj.IRKdata.fun = @impODE_GaussLegendre6;
                            obj.IRKdata.jacfun = @gauss6_jacobian;
                        case 5
                            obj.IRKdata.fun = @impODE_RadauIA_5;
                            obj.IRKdata.jacfun = @radauIA_5_jacobian;
                        case 4
                            obj.IRKdata.fun = @impODE_GaussLegendre4;
                            obj.IRKdata.jacfun = @gauss4_jacobian;
                        case 3
                            obj.IRKdata.fun = @impODE_RadauIIA_3;
                            obj.IRKdata.jacfun = @radauIIA_3_jacobian;
                        case 2
                            obj.IRKdata.fun = @impODE_ImplicitMidpoint;
                            obj.IRKdata.jacfun = @IM_jacobian;
                        case 1
                            obj.IRKdata.fun = @impODE_ImplicitEuler;
                            obj.IRKdata.jacfun = @IE_jacobian;
                    end
                        
                % TODO: not yet implemented, needs reformulation of problem
                case 'FFT' 
                    obj.fun = @obj.fft;

                case 'exact'
                    obj.fun = @obj.exp;

                % Somewhat experimental
                case {'Krylov', 'krylov'} 

                    if ~isfield(args.impODE, 'tol')
                        obj.Krylovdata.tol = 1e-6;
                    else
                        obj.Krylovdata.tol = args.impODE.tol;
                    end
                    if ~isfield(args.impODE, 'use_relative_error')
                        obj.Krylovdata.use_relative_error = false;
                    else
                        obj.Krylovdata.use_relative_error = ...
                            args.impODE.use_relative_error;
                    end
                    obj.Krylovdata.Afun = true;
                    
                    if ~isfield(args.impODE, 'rational_Krylov')
                        obj.Krylovdata.rational = false;
                    else
                        obj.Krylovdata.rational = ...
                            args.impODE.rational_Krylov;
                    end
                    
                    if obj.Krylovdata.rational
                        if ~exist('rat_Krylov')
                            warning(['The rational Krylov feature ', ...
                                     'requires that rktoolbox is ', ...
                                     'installed (in the Matlab path)']);
                        end
                        obj.fun = @obj.Krylov_rational;
                        
                        % Find best poles using surrogate optimization.
                        % This gives good results for exp(tau*A)*b for all
                        % tau in [tau_min, tau_max]. A smaller tau_min
                        % needs more iterations and poles (for some
                        % reason), and we don't know the exact interval a
                        % priori. E.g. the quadrature routine may use tau =
                        % 0 if the Clenshaw-Curtis parameters are used. But
                        % this is an outlier, and most calls will be from
                        % functions using tau in [h/10, h], so we use this
                        % interval to optimize over. At this point, h is
                        % unavailable, so the user has to provide a hint.
                        % Otherwise we use h = 1e-2.
                        if ~isfield(args.impODE, 'typical_h')
                            h = 1e-2;
                        else
                            h = args.impODE.typical_h;
                        end
                            
                        obj.Krylovdata.poles = ...
                            find_best_poles(h*1e-1, h, obj.Krylovdata.tol);
                        
                        if ~isfield(args.impODE, 'deflation_tol')
                            obj.Krylovdata.deflation_tol = 1e-10;
                        else
                            obj.Krylovdata.deflation_tol = args.impODE.deflation_tol;
                        end
                        
                    else
                        obj.fun = @obj.Krylov_polynomial;
                    end
                        
                    

                case 'Leja' % TODO: not yet functional
                    obj.fun = @obj.Leja;
                    
                case 'matlab_solver'
                    obj.fun = @obj.matlab_solver;                    

            end
            
        end

        function y = eval(obj, t0, h, z)
            y = obj.fun(t0, h, z);
        end

        function y = timevarying(obj, t0, h, z)
            r = t0;
            tend = r + h; % Note: this may be several values, h is a vector

%             Z0 = obj.M(tend(end)) \ z; % If solving for X
            Z0 = z;
            % If solving for X:
%             [y, ~, ~, ~, hfinal] = ...
%                 SDIRK_43_adaptive_timevarying(0, tend-r, ...
%                     @(s) obj.M(tend(end) - s), ...
%                     @(s) obj.A(tend(end) - s) + obj.dM(tend(end) - s), ...
%                     Z0, ...
%                     obj.IRKdata.tol, ...
%                     obj.IRKdata.hstart);
            
            [y, ~, ~, ~, hfinal] = ...
                SDIRK_43_adaptive_timevarying(0, tend-r, ...
                    @(s) obj.M(tend(end) - s), ...
                    @(s) obj.A(tend(end) - s) + obj.dM(tend(end) - s), ...
                    Z0, ...
                    obj.IRKdata.tol, ...
                    obj.IRKdata.hstart);
            % Note that we have + dM(tend-s) rather than -dM(tend-s),
            % because using tend-s rather than s changes the sign of the
            % derivative.
            
            % If solving for X:
%             Nt = length(tend);
%             if Nt == 1
%                 y = obj.M(r) * y;
%             else
%                 for j = 1:Nt
%                     y(:,(j-1)*size(Z0,2)+1:j*size(Z0,2)) = obj.M(tend(end)-h(j))* y(:,(j-1)*size(Z0,2)+1:j*size(Z0,2));
%                 end
%             end

            % Record last suggested step size (the last succesful one, i.e.
            % t(end)-t(end-1), is no good, because it was likely reduced in
            % order to exactly hit the endpoint
            obj.IRKdata.hstart = hfinal; 
        end

        function y = IRKadaptive(obj, t0, h, z)
            [y, ~, ~, ~] = SDIRK_43_adaptive(t0, t0 + h, obj.M, obj.A, z, obj.IRKdata.tol);
        end

        function y = IRK(obj, t0, h, z)
            % Ignoring t0 because it is only required in the time-varying
            % case, and this function won't work then anyway
            if ~obj.IRKdata.iterate
                k = find(obj.keys == h); % check whether we computed this matrix previously
                if isempty(k) % we didn't
                    obj.keys(end+1) = h;
                    if obj.IRKdata.disk_cache
                        J = cache_matrix(jacfun, obj.IRKdata.cachepath, h, obj.M, obj.A);
                    else
                        J = obj.IRKdata.jacfun(h, obj.M, obj.A);
                        obj.data{end+1} = J;
                    end
                    y = obj.IRKdata.fun(h, obj.M, obj.A, z, J);    
                else
                    y = obj.IRKdata.fun(h, obj.M, obj.A, z, obj.data{k});
                end
             else
                ynew = z;
                k = 1;
                rerr = inf;
                while rerr > obj.IRKdata.tol  % take k steps of size h/k with the method
                    y = ynew;
                    ynew = z;
                    hh = h/k;
                    % New J matrix
                    if obj.IRKdata.disk_cache
                        J = cache_matrix(obj.IRKdata.jacfun, obj.IRKdata.cachepath, hh, obj.M, obj.A);
                    else
                        i = find(obj.keys == hh);
                        if isempty(i)
                            J = obj.IRKdata.jacfun(hh, obj.M, obj.A);
                            obj.keys(end+1) = hh;
                            obj.data{end+1} = J;                        
                        else
                            J = obj.data{i};
                        end
                    end
                    
                    for j = 1:k
                        ynew = obj.IRKdata.fun(hh, obj.M, obj.A, ynew, J);
                    end
                    rerr = norm(ynew-y, 'fro') / norm(ynew, 'fro');
                    k = k*2;
                end
                y = ynew;
            end
            
        end
        
        function y = exp(obj, t0, h, z)
            % Ignoring t0 because it is only required in the time-varying
            % case, and this function won't work then anyway
            k = find(obj.keys == h); % check whether we computed this exponential previously
            if isempty(k) % we didn't
                ehA = expm(h*(obj.M\obj.A));
                obj.keys(end+1) = h;
                obj.data{end+1} = ehA;
                y = ehA*z;
            else
                y = obj.data{k}*z;                
            end
            
            return
        end
        
        function y = FFT(obj, t0, h, z)
            % Needs some reformulation of the problem
            N = (size(z, 1) - 1) / 2;
            shifted = fftshift(fft(z));
            for j = 1:size(shifted, 2)
                shifted(:,j) = exp(-h*((-N:N)' * 2*pi).^2) .* shifted(:,j);
            end
            y = ifft(ifftshift(shifted));
            return
        end
        
        function y = Krylov_polynomial(obj, t0, h, z)
            % Note the minus sign in the A-function... Important, as the
            % Krylov method is set up to compute expm(-A)*X
            % Ignoring t0 for now, because it is only required in the time-
            % varying case, and this function is not setup to work for that
            [y, converged, errest, kmax, subspace] = expm_block_arnoldi(@(x) -h*(obj.M\(obj.A*x)), z, obj.Krylovdata.tol, obj.Krylovdata);
%             size(subspace.Vk, 2)
%             errest
            if ~converged
                fprintf('Krylov iteration did not converge!\n')
            end
        end
        
        function y = Krylov_rational(obj, t0, h, z)
            % Ignoring t0 for now, because it is only required in the time-
            % varying case, and this function is not setup to work for that
            
            % Set up structure for rktoolbox
            obj.Krylovdata.AM.multiply = @(rho, eta, x) ...
                h*rho*(obj.M\(obj.A*x)) - h*eta*x;
            obj.Krylovdata.AM.solve = @(nu,mu,x) obj.M * (h*nu*obj.A - h*mu*obj.M)\x;
            
            
            [N, p] = size(z);
            % Repeat the poles until we have about 3 of them. According to
            % the estimate, they should be repeated about n times (where n
            % is computed in the surrogate optimization routine), but that
            % is for the worst case. We will therefore stop periodically to
            % compute an error estimate and maybe stop earlier.
            xi = repmat(obj.Krylovdata.poles, 1, ...
                        ceil(3 / length(obj.Krylovdata.poles)));
%                     xi = inf*ones(1,3);
            % Add single infinite pole for simplified formulas. (Should be
            % cheap in comparison to solving the systems connected to the
            % finite poles.)
            xi = [xi, inf]; 
            maxit = 25;
            
            % Do a first preliminary computation with a single infinite 
            % pole to set up the structures and decrease code repetition
            [V, K, H, out] = rat_krylov(obj.Krylovdata.AM, z, inf, obj.Krylovdata);
            blocksizes = out.blocksizes;
            
            obj.Krylovdata.extend = p;
            converged = false;
            for l = 1:maxit
               %%% Update Krylov subspace

                [V, K, H, out] = rat_krylov(obj.Krylovdata.AM, V, K, H, xi, obj.Krylovdata);

                % The first new blocksize will be NaN, because we are
                % extending
                blocksizes = [blocksizes, out.blocksizes(2:end)];
                r_b = [0, cumsum(blocksizes)]; % end-indices for each block
                q = blocksizes(end); % Size of final block

                Vm = V(:, r_b(1)+1:r_b(end-1));
                Vmp1 = V(:, r_b(end-1)+1:r_b(end));
%                 Hm = H(r_b(1)+1:r_b(end-1), :);
%                 Hmp1 = H(r_b(end-1)+1:r_b(end), :);
%                 Km = K(r_b(1)+1:r_b(end-1), :);
                
                %C = Hmp1/Km; % can't use this due to possible deflation
                C = Vmp1'*(obj.M\(obj.A*Vm));
                % Extend with zero rows and columns for error estimates
                C = [C; zeros(p-q, size(C,2))];
                Vmp1 = [Vmp1, zeros(N, p-q)];
                
%                 Am = Hm/Km; % can't use this due to possible deflation
                Am = Vm'*(obj.M\(obj.A*Vm));
                n = size(Am, 1);                
                
                % Ritz values
                evals = eig(Am);
                emin = min(evals);
                emax = max(evals);
                
                %%% Error estimate and approximation
                
                % Augmented matrix
                Hmt = [    Am,              zeros(n, 2*p)          ;
                            C,       emin*eye(p,p),     zeros(p, p);
                       zeros(p, n),    eye(p,p),     emax*eye(p,p)];

                % Exponentiation of this augmented matrix yields expm(h*Am)
                % as well as more useful quantities
                eHmt = expm(h*Hmt);
                ehAm = eHmt(1:n, 1:n);
                phi1AmE1 = eHmt(n+1:n+p, 1:p);
                phi2AmE1 = eHmt(n+1+p:n+2*p, 1:p);
                
                % Can't rely on Vm'*z being equal to norm(z)*eye(...) 
                % because of permutations
                y = Vm*ehAm*(Vm'*z);
                
                % Coarse estimate which seems to be too coarse
%                 errest1 = norm(z)*norm(Vmp1*phi1HmE1);
                
                % Better estimate, which requires one multiplication with A
                e1 = Vmp1*phi1AmE1;
                e2 = obj.M\(obj.A*Vmp1*phi2AmE1);
                e3 = emin*Vmp1*phi2AmE1;
                errest = norm(z)*norm(e1 + e2 + e3);
                
                if errest < obj.Krylovdata.tol
                    converged = true;
%                     errest
                    break
                end

            end
            if ~converged
                fprintf('Krylov iteration did not converge!\n')
            end
        end
        
        function y = Leja(obj, t0, h, z)
        end
        
        
        function y = matlab_solver(obj, t0, h, z)
            
            f = @(t, T) obj.A*T;
            jac = @(t,y) obj.A;
%             opts = odeset('AbsTol', 1e-2, 'RelTol', 1e-4, 'Mass', obj.M, 'OutputFcn', @odeprint, 'OutputSel', 1, 'Jacobian', jac);
            opts = odeset('AbsTol', 1e-2, 'RelTol', 1e-4, 'Mass', obj.M, 'MassSingular', false, 'Jacobian', jac);
            y = zeros(size(z));
            for k = 1:size(z, 2)
                [t, yc] = ode15s(f, [t0, t0 + h], z(:,k), opts);
                y(:,k) = yc(end,:)';
                fprintf('%d of %d done\n', k, size(z, 2))
            end
            fprintf('Done\n')
        end
    end % methods
    
end % classdef
